;;;; The XML-in-LAML common parts, which are shared between all XML-in-LAML languages.
;;;; This library is designed to co-exists with non-validating HTML mirrors (for the sake of LENO, for instance).
;;;; It cannot be used together with validating ast-based HTML mirrors, such as html4.01-transitional-validating.
;;;; If you need to, you should use an all xml-in-laml solution. By that we mean a solution, where also the
;;;; HTML stuff is based on XML-in-LAML. In practical terms, it means that you should use the XML-in-LAML XHTML mirror in
;;;; lib/xml-in-laml/mirrors/ 


;;; This section contains the XML declaration

;; Return the xml declaration
(define (xml-declaration)
  "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>")

;;; Configuration part.
;;; The variables and functions in this part controls the checking, validation, and error reporting of the XML mirror functions.

;; A boolean variable which controls the checking of tag attributes.
;; If true, check the tag (element) attributes against the attributes as defined in the DTD file.
;; The default value is true.
(define xml-check-attributes? #t)

;; A boolean variable that controls the XML validation.
;; If true, do validate the generated XML document against the DTD.
;; The default value is true.
(define xml-validate-contents? #t)

;; A boolean variable that controls the check of XML language overlap.
;; If true, check that that no elements are used via ambiguous simple names.
(define xml-check-language-overlap? #t)

;; A boolean variable that controls whether CDATA (Character data) are transformed
;; through the HTML/XML character transformation table. The default value is true.
;; .reference "Character transformation table" "table" "../../../man/laml.html#html-char-transformation-table"
;; !!! Should be language specific.
(define xml-transliterate-character-data? #t)

;; A list of elements for which content data - CDATA - is never transformed.
;; As an example, it will often cause problems if scripts are being modified by means of transliterations using
;; the HTML character transformation table.
;; A list of strings.
;; Defaults to the empty list
;; You can redefine it after the document style has been loaded.
;; XML language specific.
(define xml-non-transliteration-elements (list ))

;; How many characters of the validation error messages to be presented.
(define xml-error-truncation-length 130)


;;; Other Constants.


(define explicit-space #t)
(define explicit-space-suppress #f)

;; The symbolic name of explicit white space suppress
(define _ explicit-space-suppress)


; ---------------------------------------------------------------------------------------------------
;;; AST functions.
;;; In this section we document the functions that work on abstract syntax trees.
;;; All the XML mirror functions return abstract syntax trees, which need to be transformed
;;; to text before they can be presented. The function xml-render performs this transformation.

;; A boolean variable that controls whether or not we check the internal soundness of ASTs.
;; If the basic mirror is used directly, it makes good sense to set this variable to #t.
;; Else we recommend a #f value.
(define check-ast-constituents? #f)

;; Make an AST given element-name, contents, a property list of attributes, a kind (single/double)
;; an an XML language name. 
;; elment-name must be a string or symbol.
;; The parameter contents may be a list of subtrees. It may also be or an element which satisfies
;; either ast?, cdata? or forced-white-space?. In these cases, we embed the single element into a list.
;; attributes is the list of attributes on property list format (such as '(a1 "v1" a2 "v2")). 
;; attribute names are always symbols, and attribute values are always strings.
;; Kind is either the symbol double or single.
;; Language is a symbol or string that represent the XML language name.
(define (make-ast element-name contents attributes kind language)
 (let ((subtrees (cond ((ast? contents) (list contents))
                       ((cdata? contents) (list contents))
                       ((forced-white-space? contents) (list contents))
                       ((list? contents) contents)
                       (else (laml-error "make-ast: Contents must be ast, cdata, forced white space, or a list of these: " 
                                         (as-string contents))))))
  (list 'ast (as-string element-name) subtrees attributes (as-symbol kind) (as-symbol language))))

; AST selectors.

;; Return the root element name of ast.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
(define (ast-element-name ast)
  (cadr ast))

;; Return the list of substrees of ast.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
(define (ast-subtrees ast)
  (caddr ast))

;; Return the list of attributes of ast, on property list format.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
(define (ast-attributes ast)
  (cadddr ast))

;; Return the kind of the ast. It is either the symbol single or double.
;; Tells whether the ast is to be rendered as a single or double tag.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
(define (ast-kind ast)
  (fifth ast))

;; Return the language of teh ast (a symbol).
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
(define (ast-language ast)
  (sixth ast))


; AST predicates

;; Is the parameter x an AST
(define (ast? x)
  (and (pair? x) (eq? (car x) 'ast)
       (list? x) (= 6 (length x))))

;; Is the parameter x CDATA
;; .form (cdata? x)
(define cdata? string?)

;; Is the parameter x either an ast or a contents string.
(define (contents-data? x)
 (or (cdata? x) (ast? x)))

;; Is the parameter x a white space mark.
(define (forced-white-space? x)
 (and (boolean? x) x))

;; Is the parameter x white space related?
;; .form (white-space-related? x)
(define white-space-related? boolean?)


; ---------------------------------------------------------------------------------------------------
;;; XML language registration and language related bookkeeping.


;; The list of names (symbols) which causes name clashes relative to 
;; the current set of languages in use.
;; Assigned by register-xml-in-laml-language.
;; Thus, this variable is always up-to-date in between registrations of languages.
;; Be careful not to redefine this variable by double loading this file.
(define xml-in-laml-name-clashes '())

;; Register that language (the first parameter) is an XML language in use in the current LAML session.
;; The first parameter, language, is a symbol.
;; The second parameter, language-map, is the language map of the language.
;; As a precondition is a assumed that we do not register the same language more than once.
(define (register-xml-in-laml-language language language-map)
 (set! xml-in-laml-name-clashes (precompute-name-clashes language-map xml-in-laml-name-clashes)) 
 (if (not (memq language (languages-in-use)))
     (set! xml-in-laml-languages-in-use (cons (cons language language-map) xml-in-laml-languages-in-use)))
)

; Return the new name clashes given that new-language-map is about to be registered.
; The existing name clashes is passed as the parameter existing-name-clashes.
(define (precompute-name-clashes new-language-map existing-name-clashes)
  (let* ((all-existing-names (flatten (map element-names-of-language (languages-in-use))))
         (new-names (map car new-language-map))
         (intersection (list-intersection-by-predicate new-names all-existing-names eq?))
        )
    (remove-duplicates-by-predicate (append existing-name-clashes intersection) eq?)))


;; Return the language map of language.
;; If language is not registered via register-xml-in-laml-language, return #f.
;; The parameter language is a symbol.
(define (language-map-of language)
  (defaulted-get language xml-in-laml-languages-in-use #f))

;; is language in use - is it registered as an xml-in-laml language?
(define (language-in-use? language)
  (let ((lg-map (defaulted-get language xml-in-laml-languages-in-use #f)))
    (if lg-map #t #f)))

;; Return the list of language in used, as registered by register-xml-in-laml-language.
(define (languages-in-use) 
  (map car xml-in-laml-languages-in-use))

;; Return the element names of language, as defined by its language map.
(define (element-names-of-language language)
  (let ((lg-map (defaulted-get language xml-in-laml-languages-in-use '())))
    (map car lg-map))) 

;; Is name involed in a name clash among the registered xml-in-laml languages?
(define (causes-xml-in-laml-name-clash? name)
  (memq name xml-in-laml-name-clashes))

(define (activator-via-language-map language)
 (lambda (element-name)
  (let ((lg-map (language-map-of language)))
    (get-mirror-function lg-map element-name))))



; ---------------------------------------------------------------------------------------------------
;;; The language map.
;;; The language map is a mapping that allows us to access the element mirror functions independent of
;;; any name clash. 
;;; In this version the language map is an association list from element names (symbols) to function objects.
;;; In the longer run, it should be more efficient search structure.

; An auxilary variable used by the xml-in-laml tool:
(define temp-language-map '())

; An auxiliry variable used by the xml-in-laml tool:
(define temp-mirror-function #f)

;; Return the mirror function based on element-name from language map.
;; Return #f if no mirror function exists in the map.
;; The parameter element-name is either a symbol or a string.
(define (get-mirror-function language-map element-name)
 (let ((element-name-symbol (as-symbol element-name)))
   (defaulted-get element-name-symbol language-map #f)))

;; Update the language map by the association of element-name and mirror function.
;; Returns the updated map.
;; The parameter element-name is either a symbol or a string.
;; Causes an error if there already is an entry for element-name
(define (put-mirror-function language-map element-name mirror-function)
 (let ((element-name-symbol (as-symbol element-name)))
   (if (get-mirror-function language-map element-name-symbol)
       (laml-error "put-mirror-function: The name" element-name "is defined twice.")
       (cons (cons element-name-symbol mirror-function) language-map))))


; -----------------------------------------------------------------------------------------------------

;;; Mirror generation functions.
;;; The functions in this sections are the important and 'heavy' mirror generation functions.

;; The central function which 'sorts out' the parameters of the mirror functions.
;; Returns a cons cell of contents and attributes.
(define (xml-sort-tag-parameters parameters tag-name)
 (xml-sort-tag-parameters-1 parameters parameters '() '() tag-name))

(define (xml-sort-tag-parameters-1 original-parameters parameters contents attributes tag-name)
 (letrec ((strip-initial-explicit-spaces 
	   (lambda (cl) (if (and (pair? cl) (eq? (car cl) explicit-space)) (strip-initial-explicit-spaces (cdr cl)) cl)))
	  (list-non-ast? (lambda (x) (and (list? x) (not (ast? x)))))
         )
  (cond ((null? parameters) (cons (reverse (strip-initial-explicit-spaces contents)) (reverse attributes)))

        ; CONTENTS DATA -  no space after
        ((and (contents-data? (car parameters)) (not (null? (cdr parameters)))
              (boolean? (cadr parameters)) (not (cadr parameters)))              
                 (xml-sort-tag-parameters-1 original-parameters (cddr parameters) (cons (car parameters) contents) attributes tag-name))

        ; CONTENTS DATA -  space after
        ((and (contents-data? (car parameters)) (not (null? (cdr parameters)))
              (boolean? (cadr parameters)) (cadr parameters))                    ; space after
                 (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) (cons explicit-space (cons (car parameters) contents)) attributes tag-name))

        ; CONTENTS DATA -  space after
        ((contents-data? (car parameters))   ; space after
                 (xml-sort-tag-parameters-1 original-parameters 
                                        (cdr parameters) (cons explicit-space (cons (car parameters) contents)) attributes tag-name))


        ; LIST - no space after
        ((and (list-non-ast? (car parameters)) (not (null? (cdr parameters))) (boolean? (cadr parameters)) (not (cadr parameters)))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 )  
              (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) 
                                        (append (reverse rec-contents) contents) (append (reverse rec-attributes) attributes) tag-name)))

        ; LIST - space after
        ((and (list-non-ast? (car parameters)) (not (null? (cdr parameters))) (boolean? (cadr parameters))  (cadr parameters))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 )  
              (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) 
                                        (append (list explicit-space) (reverse rec-contents) contents) 
                                        (append (reverse rec-attributes) attributes) tag-name)))

        ; LIST - space after
        ((and (list-non-ast? (car parameters)))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 ) 
              (xml-sort-tag-parameters-1 original-parameters
                                        (cdr parameters) 
                                        (append (list explicit-space)
                                                (reverse rec-contents)
                                                contents)
                                        (append (reverse rec-attributes) attributes) tag-name)))

        ; SYMBOL - attribute name 
        ((and (symbol? (car parameters)) (not (null? (cdr parameters))) (not (boolean? (cadr parameters))))
           (let ((attr-name (car parameters))
                 (attr-val (cadr parameters)))
             (xml-sort-tag-parameters-1 original-parameters (cddr parameters) contents (cons attr-val (cons attr-name attributes)) tag-name)))

        ; SYMBOL - error
        ((symbol? (car parameters)) (xml-sort-error (string-append "Attributes of the " tag-name
                                    " tag must be of the form 'symbol \"value\". Only the symbol " (as-string (car parameters))
                                    " appears in last attribute") original-parameters))

        (else (xml-sort-error (string-append "Error in an LAML " tag-name " tag.") original-parameters))
  )))

(define (xml-sort-error message parameters)
 (let* ((max-str-lgt 80)  ; symbolic constant definition
        (parameters-1 (as-quoted-string parameters))
        (parameters-2 (if (> (string-length parameters-1) max-str-lgt) 
                        (string-append (substring parameters-1 0 (- max-str-lgt 1)) "...")
                        parameters-1)))
  (error
    (string-append
      message (as-string #\newline)
      "  FORM CONTENTS AND ATTR.: " parameters-2 (as-string #\newline)))))

; -----------------------------------------------------------------------------------------------------------------------

;; Return an XML mirror surface function, in which textual content parameters and attribute
;; value pairs can be given in a very liberal fashion.
;; The parameter validation-procedure is the procedure that validates all aspects of the application of the element mirror function.
;; The parameter tag-name is a string which represents the name of the tag (used for error message purposes).
;; The paramter single-double-kind is either the symbol single, double or possible-single.
;; The paramter language defines the language in use (a symbol).
;; The parameter overlap-check? controls whether to check for name clashes in between the registered XML-in-LAML languages.
;; The parameter action-procedure serves both as a boolean and a procedure.
;;   If non-false, the parameter servers as action procedure, which will receive the AST for action purposes.
(define (generate-xml-mirror-function validation-procedure tag-name single-double-kind language overlap-check? action-procedure)
 (cond ((eq? single-double-kind 'double)
	(lambda parameters
	  (let* ((contents-attributes (xml-sort-tag-parameters parameters tag-name))
		 (contents (car contents-attributes))
		 (attributes (cdr contents-attributes))
		)
	    (validation-procedure tag-name attributes contents overlap-check?)
	    (if action-procedure
                (let ((the-ast (make-ast tag-name contents attributes 'double language)))
                  (action-procedure the-ast)
                  the-ast)
                (make-ast tag-name contents attributes 'double language)
            )
	  )
	)
       )

       ((eq? single-double-kind 'single)   
          (lambda parameters
            (let* ((contents-attributes (xml-sort-tag-parameters parameters tag-name))
		   (contents (car contents-attributes))
		   (attributes (cdr contents-attributes))
		  )
	      (validation-procedure tag-name attributes contents overlap-check?)
              (if action-procedure
                  (let ((the-ast (make-ast tag-name '() attributes 'single language)))
                    (action-procedure the-ast)
                    the-ast)
	          (make-ast tag-name '() attributes 'single language)
              )
            )
          )
       )

       (else (error (string-append "generate-xml-mirror-function: unknown single-double-kind: " (as-string single-double-kind))))
 )
)



; ---------------------------------------------------------------------------------------------------
;;; Contents validation. 
;;; The definitions in this section is used by the validation procedures, which are specific
;;; for each supported XML language

; The list on which we accumulate encountered validation problems.
(define xml-problem-contents-list '())

; The last element which caused a problem in xml-member-remembe
(define xml-problem-element #f)

; A variant of member which registers context on xml-problem-contents-list in case (not (member el lst))
(define (xml-member-remember el lst context)
 (let ((res (member el lst)))
   (if (not res)
       (begin
         (set! xml-problem-contents-list (cons context xml-problem-contents-list))
         (set! xml-problem-element el)
       ))
   res))

;; Return a predicate which checks whether the contents matches one or more instances from element-constituent-list.
;; Element constituent-list is a list of strings, namely element tag names or
;; "#PCDATA".
;; A true results means OK - no error
;; Avoid a O(n^2) algorithm here in the final implementation
(define (one-or-more . element-constituent-list)
  (lambda (contents)
    (set! xml-problem-contents-list '())
    (cond ((or (and (boolean? contents) (not contents))
               (and (list? contents) (null? contents))) 
             (begin
                  (set! xml-problem-contents-list (cons "Empty contents not acceptable" xml-problem-contents-list))
                  #f))  ; empty contents not acceptable
          ((ast? contents) (xml-member-remember (ast-element-name contents) element-constituent-list contents))
	  ((cdata? contents) (xml-member-remember "#pcdata" element-constituent-list contents))
	  ((forced-white-space? contents) #t)
	  ((list contents)
            (and 
	     (>= (length contents) 1)	; implement more efficienly
	     (accumulate-right
	      (lambda (x y) (and x y))
	      #t
	      (map 
	       (lambda (cont)
		 (cond ((and (boolean? cont) (not cont))
                         (begin
			   (set! xml-problem-contents-list (cons "Empty contents not acceptable" xml-problem-contents-list))
			   #f)) ; even a single #f is not acceptable
                       ((ast? cont) (xml-member-remember (ast-element-name cont) element-constituent-list cont))
		       ((cdata? cont) (xml-member-remember "#pcdata" element-constituent-list cont))
		       ((forced-white-space? cont) #t)
		       (else #f)))
	       contents)))
          )
	  (else (laml-error "one-or-more: Unknown contents: " (as-string contents))))))

;; Return a predicate which checks whether the contents matches zero or more instances from element-constituent-list
(define (zero-or-more . element-constituent-list)
  (lambda (contents)
    (set! xml-problem-contents-list '())
    (cond ((and (boolean? contents) (not contents)) #t)  ; no contents, which is acceptable
          ((ast? contents) (xml-member-remember (ast-element-name contents) element-constituent-list contents))
	  ((cdata? contents) (xml-member-remember "#pcdata" element-constituent-list contents))
	  ((forced-white-space? contents) #t)
	  ((list contents)
	    (accumulate-right
	     (lambda (x y) (and x y))
	     #t
	     (map 
	      (lambda (cont)
		(cond ((and (boolean? cont) (not cont)) #t) ; list no contents, which is acceptable
                      ((ast? cont) (xml-member-remember (ast-element-name cont) element-constituent-list cont))
		      ((cdata? cont) (xml-member-remember "#pcdata" element-constituent-list contents))
		      ((forced-white-space? cont) #t)
		      (else #f)))
	      contents))
	   )
	  (else (laml-error "one-or-more: Unknown contents: " (as-string contents))))))



;; Return a predicate which checks whether the contents match the sequential rhs-elements.
;; rhs-elements is a list of strings or symbols, which symbols represent optionals.
;; Example (sequence-with-optionals "xxx" "yyy" 'zzz) corresponds to
;;         xxx, yyy, zzz?
;; We assume that rhs-elements, in the starting point, always contains at least one element.
(define (sequence-with-optionals . rhs-elements)
  (lambda (contents)
    (set! xml-problem-contents-list '())
    (cond ((or (and (boolean? contents) (not contents))
               (and (list? contents) (null? contents))) 
             (xml-add-problem! (xml-enrich-error-message "Empty contents not acceptable in sequence" contents))
             ) 
	  ((cdata? contents) (xml-add-problem! (xml-enrich-error-message "PCDATA not acceptable in this context" contents)))
	  ((forced-white-space? contents) (xml-add-problem! "Empty contents not acceptable in sequence"))
	  ((list contents)
            (and 
	     (>= (length contents) 1)	
             (sequence-with-optionals-1 contents rhs-elements)
	    )
          )
	  (else (laml-error "sequence-with-optionals: Unknown contents: " (as-string contents))))))

; A tail recursive helping function to sequence-with-optionals.
; contents-list is a list contents elements.
; rhs-elements is a list of sequential elements of the parsed rhs (strings or symbol lists).
(define (sequence-with-optionals-1 contents-list rhs-elements)
  (cond ((and (null? contents-list) (null? rhs-elements)) #t) ; OK
        ((and (null? contents-list) (remaining-optionals? rhs-elements)) #t) ; content-list end - one optional rhs el. OK.
        ((null? rhs-elements) (xml-add-problem! (xml-enrich-error-message "Sequence too long" contents-list)))
        ((null? contents-list) (xml-add-problem! (xml-enrich-error-message "Sequence too short" contents-list)))

        ; now both lists non-empty

        ((cdata? (car contents-list))
            (xml-add-problem! (xml-enrich-error-message "Encountered PCDATA where sequence is expected" (string-it (car contents-list)))))

        ((forced-white-space? (car contents-list)) (sequence-with-optionals-1 (cdr contents-list) rhs-elements))

        ((and (ast? (car contents-list)) (string? (car rhs-elements)) 
              (equal? (ast-element-name  (car contents-list)) (car rhs-elements)))  ; two matching elements - proceed booth
           (sequence-with-optionals-1 (cdr contents-list) (cdr rhs-elements)))

        ((and (forced-white-space? (car contents-list)))  ; forced white space - proceed
           (sequence-with-optionals-1 (cdr contents-list) rhs-elements))

        ((and (ast? (car contents-list)) (symbol? (car rhs-elements)) 
              (equal? (ast-element-name  (car contents-list)) (as-string (car rhs-elements)))) ; match against optional - proceed booth
           (sequence-with-optionals-1 (cdr contents-list) (cdr rhs-elements)))

        ((and (ast? (car contents-list)) (symbol? (car rhs-elements)) 
              (not (equal? (ast-element-name  (car contents-list)) (as-string (car rhs-elements))))) ; non match against optional - proceed rhs only
           (sequence-with-optionals-1 contents-list (cdr rhs-elements)))

        (else  ; all other cases cause mismatch
           (xml-add-problem! (xml-enrich-error-message "Mismatch in sequence" contents-list) ))))

(define (remaining-optionals? lst)
  (cond ((null? lst) #t)
        (else (and (symbol? (car lst)) (remaining-optionals? (cdr lst))))
  )
)

;; A predicate which checks whether the contents is PCDATA.
(define (pcdata-checker contents)
 (set! xml-problem-contents-list '())
 (let ((contents-1 (if (list? contents) (filter (negate white-space-related?) contents) contents)))
  (cond ((and (list? contents) (null? contents))
          (begin
	   (set! xml-problem-contents-list 
		 (cons (xml-enrich-error-message "The instance should not be empty" contents-1)
		       xml-problem-contents-list))
	   #f))
        ((cdata? contents-1)		; #f - no contents probably
           #t)
        ((not (list? contents-1))
          (begin
	   (set! xml-problem-contents-list 
		 (cons (xml-enrich-error-message "The instance should not be empty" contents-1)
		       xml-problem-contents-list))
	   #f))
	((list? contents-1)
	 (if (not
	       (accumulate-right and-fn #t (map cdata? contents-1))) ; all cdata?
	     (begin
	       (set! xml-problem-contents-list 
		     (cons (xml-enrich-error-message 
                             "The instance must consist of textual contents - no embedded elements allowed" contents-1)
			   xml-problem-contents-list))
	       #f)
	     #t))
	(else #f))))


; Adds error message to xml-problem-contents-list and returns #f, as an indication of validation failure.
(define (xml-add-problem! error-message)
 (set! xml-problem-contents-list (cons error-message xml-problem-contents-list))
 #f)

; Enrich an error message with info about a constitutent (AST)
(define (xml-enrich-error-message error-message constituent)
  (string-append error-message ":   "  (xml-render-1 constituent)))

; Enrich a list of error messages with info about a constitutent (AST)
(define (xml-enrich-error-messages error-message-list constituent)
  (if (list? error-message-list)
      (map (lambda (e) (string-append e ":   " (xml-render-1 constituent))) error-message-list)
      error-message-list))

; and as a function - as opposed to a macro.
(define (and-fn x y) (and x y))


; Is there an AST with el-name in the list of ASTs.
; precondition: contents is a list of ASTs
(define (there-exists-ast? el-name contents-list)
  (find-in-list (lambda (ast) (equal? (ast-element-name ast) el-name)) contents-list))

; Return the instances of ast element name in the contents list.
; precondition: contents is a list of ASTs
(define (instances-of-ast el-name contents-list)
  (filter (lambda (ast) (equal? (ast-element-name ast) el-name)) contents-list))


; Do all ASTs have element names in el-name-list
; precondition: contents is a list of ASTs
; Should not be O(n^2) in a more finished implementation
(define (all-ast-member-of? el-name-list contents-list)
 (accumulate-right and-fn #t 
  (map (lambda (ast) (member (ast-element-name ast) el-name-list)) contents-list)))

; A version of filter which work normally on lists, and else just returns the empty list.
(define (filter-contents pred contents-list)
 (if (list? contents-list)
     (filter pred contents-list)
     '()))

; ---------------------------------------------------------------------------------------------------------------

; Apply error-predicate on contents. If the validation-error-predicate returns a true result,
; issue a warning. The result is suposed to give a hint about the actual error. 
; Rewrap it!!!
(define (validate-contents! contents validation-predicate tag-name)
 (set! xml-problem-contents-list '())
 (let ((validation-status (validation-predicate contents)))
  (if (not validation-status)
      (let ((count (length xml-problem-contents-list)))
        (if (> count 0) ; should always be true
          (xml-check-error
           (string-append 
            (string-append 
              (if (= count 1) 
		  "XML validation error(s) encountered in an instance of the  "
		  (string-append  "XML validation error(s) encountered in an instance of the  "))
              (as-string tag-name) "  element")
            (indented-terminal-lines (map (compose truncate-string xml-render-1) xml-problem-contents-list)))))))))

(define (indented-terminal-lines line-list)
 (let ((sep (string-append (as-string #\newline) "  ")))
  (string-append sep (list-to-string line-list sep))))

(define (xml-render-1 contents)
  (cond ((ast? contents) (xml-render contents))
	((cdata? contents) contents)
	((forced-white-space? contents) "SPACE")
	((list? contents) (list-to-string (map xml-render-1 contents) " "))
        (else "??")))

(define (truncate-string str)
 (if (> (string-length str) xml-error-truncation-length)
     (string-append (substring str 0 xml-error-truncation-length) "...")
     str))

; A predicate which always gives positive validation.
; Used if the xml-in-laml tool constant use-manually-programmed-validation-predicates is #f.
(define (faked-generice-content-checker x)
  #t)

; ------------------------------------------------------------------------------------------------------------------------------
; Empty contents checking of empty elements

(define (xml-check-for-empty-contents! contents tag-name)
  (if (not (null? contents))
      (xml-check-error 
        (string-append "The single tag " tag-name " is not supposed to have any contents: "
                       (xml-render-1 contents) ". We just skip it."))))


; --------------------------------------------------------------------------------------------------------------------------------
;;; Attribute checking. 
;;; The definitions in this section is used by the validation procedures, which are specific
;;; for each supported XML language 

; Attribute tripple selectors
(define att-name (make-selector-function 1))
(define att-type (make-selector-function 2))
(define att-status (make-selector-function 3))

;; Check the attributes (first par) in the calling form against the attribute definitions taken from the DTD file (second par).
;; The first parameter, attribute, is the attributes of the calling form, on property list form:
;; (a1 v1 ... an vn). ai is a symbol and vi is a string.
;; The second parameter dtd-attribute-definition is the attributes as defined
;; in the dtd for the element in question. The third parameter number-of-req-attributes is the number
;; of required attributes in dtd-attr-definitions. The fourth parameter, tag-name, is the
;; name of the enclosing tag. dtd-attribute-definition is a list of 
;; triples (attr-name attr-type require-status). attr-name is a string, attr-type is
;; a string or a list of strings (possibilities), and require-status is a string
;; such as "#IMPLIED" or "#REQUIRED"
(define (xml-check-attributes! attributes dtd-attribute-definition number-of-req-attributes tag-name)
 (if (even? (length attributes))  ; if not, we report the error elsewhere
  (let ((required-attribute-names 
           (map (compose as-symbol car) (front-sublist dtd-attribute-definition number-of-req-attributes)))
        (dtd-attribute-names (map (compose as-symbol car) dtd-attribute-definition))
        (attribute-names (if (null? attributes) '() (every-second-element attributes)))
        (attribute-values (if (null? attributes) '() (every-second-element (cdr attributes))))
       )
    (xml-check-required-attributes! attribute-names required-attribute-names tag-name)
    (xml-check-for-attribute-existence! attribute-names dtd-attribute-names tag-name)
    (xml-check-for-attribute-types! attribute-names attribute-values dtd-attribute-definition tag-name)
)))


; the symbols in required-attribute-names must all be found as attribute names in the property list attributes
; if not an error is issued.
; tag-name is the name of the enclosing tag
(define (xml-check-required-attributes! attribute-names required-attribute-names tag-name)
  (if (not (null? required-attribute-names))
      (begin
         (xml-check-one-required-attribute! attribute-names (car required-attribute-names) tag-name)
         (xml-check-required-attributes! attribute-names (cdr required-attribute-names) tag-name))))

(define (xml-check-one-required-attribute! attribute-names required-attribute tag-name)
  (if (not (memq required-attribute attribute-names))
      (xml-check-error (string-append "The required attribute  " (as-string required-attribute) "  is not present in the " tag-name " element."))))

; A variant of display-warning which prefixes the warning text with 'XML'.
(define (display-xml-warning message)
  (display (string-append "XML Warning: " message)) (newline))


;; The definition of the procedure that reports validation errors.
;; The default value is display-xml-warning, which issues non-fatal warning messages.
;; Alternative value: laml-error (fatal error messaging).
(define xml-check-error display-xml-warning)


; Are all names in attribute-names found in the list of names dtd-attribute-names
; tag-name is the name of the enclosing tag
; CSS attributes should not be checked for existence. 
(define (xml-check-for-attribute-existence! attribute-names dtd-attribute-names tag-name)
   (if (not (null? attribute-names))
       (begin
         (xml-check-one-attribute-existence! (car attribute-names) dtd-attribute-names tag-name)
         (xml-check-for-attribute-existence! (cdr attribute-names) dtd-attribute-names tag-name))))

(define (xml-check-one-attribute-existence! name dtd-attribute-names tag-name)
  (if (and (not (xml-css-key? name)) (not (memq name dtd-attribute-names)))
      (xml-check-error (string-append "The XML attribute  " (as-string name) "  is not valid in the "  tag-name " element."))))

; Check that the attributes (as splitted into attribute-names and attribute-values - list of equal lengths)
; have correct types accoring the dtd-attribute-names.
(define (xml-check-for-attribute-types! attribute-names attribute-values dtd-attributes tag-name)
  (if (not (null? attribute-names))
      (begin
         (xml-check-one-attribute-type! (car attribute-names) (car attribute-values) dtd-attributes tag-name)
         (xml-check-for-attribute-types! (cdr attribute-names) (cdr attribute-values) dtd-attributes tag-name))))

(define (xml-check-one-attribute-type! name value dtd-attributes tag-name)
  (let ((attribute-descriptor 
          (find-in-list (lambda (tripple) (eq? (as-symbol (car tripple)) name)) dtd-attributes)))
    (if attribute-descriptor
        (xml-check-attribute-value! name value (att-type attribute-descriptor) tag-name))))

; is value a 'member of' attribute-type. value is a string.
; Attribute is a string such as "NUMBER" "CDATA" or a list of strings, in which case value must be a member.
(define (xml-check-attribute-value! name value attribute-type tag-name)
  (cond ((list? attribute-type) 
            (if (not (member (downcase-string value) (map downcase-string attribute-type)))  ; too expensive to downcase list here 
                (xml-check-error (string-append "The value  " (string-it value) "  of the XML attribute  "
                                            (as-string name) "  is not valid in the "  tag-name " element."))))
        ((equal? "NUMBER" attribute-type)
            (if (not (numeric-string? value))
                (xml-check-error (string-append "The attribute value  " (string-it value) "  should be numeric (but enclosed in string quotes)"))))
        (else 'incomplete)  ; not implemented yet
  ))


; ---------------------------------------------------------------------------------------------------------------
; Check of language overlap

(define (check-language-overlap! name)
  (if (causes-xml-in-laml-name-clash? name)
      (xml-check-error 
       (string-append "The mirror function named " (as-string name) " is ambiguous. Please use it via an appropriate language map."))))



;;; Rendering functions and procedures.
;;; The functions and procedures in this section transform the internal document representation
;;; (abstract syntax trees) to a textual form (text strings).

; ---------------------------------------------------------------------------------------------------
; Fast AST rendering.

;; Render the XML clause (an AST) to output-port.
;; output-port is assumed to be open. output-port will not be closed by this procedure.
;; .form (render-to-output xml-clause output-port [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter output-port an open output port
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
(define (render-to-output-port xml-clause output-port . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
        (put-fn (put-in-sink-stream-generator output-port))
       )
  (put-fn prolog-text)
  (render-fast xml-clause put-fn)
  (put-fn epilog-text)))

;; Render the xml-clause (an AST) and return the rendered string
;; .form (render xml-clause [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
(define (xml-render xml-clause . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
       )
   (reset-sink-string)
   (render-fast xml-clause put-in-sink-text-string)
   (string-append
     prolog-text
     (sink-string)  
     epilog-text)))



; --------------------------------------------------------------------
; String sink in terms of segments (a list) of strings.

(define sink-string-segment-size 20000)
(define sink-string-segment-limit (- sink-string-segment-size 1))
(define output-sink-segments '())
(define output-sink-string (make-string sink-string-segment-size))
(define next-sink-ptr 0)

; Put x (string or char) into output sink.
(define (put-in-sink-text-string x)
  (cond ((char? x) 
          (begin
	   (string-set! output-sink-string next-sink-ptr x)
	   (set! next-sink-ptr (+ 1 next-sink-ptr))
           (if (= next-sink-ptr sink-string-segment-size) (prepare-next-sink-segment))))

        ((string? x)
          (do ((lgt (string-length x))
	       (i 0 (+ i 1))
               (j next-sink-ptr (if (< j sink-string-segment-limit) (+ j 1) 0))
              )
	      ((= i lgt) (set! next-sink-ptr j))
	       (string-set! output-sink-string j (string-ref x i))
               (if (= j sink-string-segment-limit) (prepare-next-sink-segment))
              ))

        (else "put-in-sink-string: Can only output chars or strings")))

(define (prepare-next-sink-segment)
 (set! output-sink-segments (cons (string-copy output-sink-string) output-sink-segments))
 (set! next-sink-ptr 0)
)

; Return the resulting sink string
(define (sink-string)
 (string-append 
   (list-to-string (reverse output-sink-segments) "")
   (substring output-sink-string 0 next-sink-ptr)))

(define (reset-sink-string)
 (set! output-sink-segments '())
 (set! next-sink-ptr 0))

; --------------------------------------------------------------------
; Stream sink.

; Generate an output port put procedure
(define (put-in-sink-stream-generator port)
 (lambda (x)
  (cond ((char? x) 
          (write-char x port))
        ((string? x)
          (do ((lgt (string-length x))
               (i 0 (+ i 1))
              )
	      ((= i lgt) 'done)
  	       (write-char (string-ref x i) port)))
        (else "put-in-sink-stream: Can only output chars or strings"))))


; --------------------------------------------------------------------
; Sink independet rendering stuff. 
; Depends on xml-linearize-attributes and html-char-transform (from laml.scm).

(define (render-fast ast put)
 (let* ((tag-name (ast-element-name ast)) 
        (contents-list (ast-subtrees ast))
        (attribute-properlist (ast-attributes ast))
        (attributes-res (xml-linearize-attributes attribute-properlist))
        (xml-attributes-string (car attributes-res))
        (css-attributes-string (cdr attributes-res))
        (kind (ast-kind ast))
       )
  (cond ((eq? kind 'single)
          (put #\<) 
          (put tag-name) 
          (if (not (null? attribute-properlist)) 
	      (begin
	       (put #\space)
	       (put xml-attributes-string)
	       (if (not (empty-string? css-attributes-string))
		   (begin (put " style = ")
			  (put #\") (put css-attributes-string) (put #\")))
	       ))
          (put "/>"))

        ((eq? kind 'double)
          (put #\<) 
          (put tag-name) 
          (if (not (null? attribute-properlist)) 
	      (begin
	       (put #\space)
	       (put xml-attributes-string)
	       (if (not (empty-string? css-attributes-string))
		   (begin (put " style = ")
			  (put #\") (put css-attributes-string) (put #\")))
	       ))
          (put #\>)
          (linearize-contents-list-fast contents-list put tag-name)
          (put "</")  (put tag-name) (put #\>))

        (error "render-fast: Either a single or double kind of ast expected."))))

(define (linearize-contents-list-fast contents-list put tag-name)
  (for-each 
    (lambda (contents) (linearize-contents-fast contents put tag-name))
    contents-list))

(define (linearize-contents-fast contents put tag-name)
  (cond ((cdata? contents) 
          (if (and xml-transliterate-character-data? (not (member tag-name xml-non-transliteration-elements)))
              (do ((lgt (string-length contents))
		   (i 0 (+ i 1))
		   )
		  ((= i lgt) 'done)
		(put (html-char-transform (string-ref contents i))))
              (put contents)))
        ((forced-white-space? contents) (put #\space))
        ((ast? contents) (render-fast contents put))))


; ------------------------------------------------------------------------------------------
; Attribute linearization. 

; Return a pair of strings: The head is the string represeting XML attributes.
; The tail is the string representing the CSS attributes (to go into an HTML style attribute).
(define (xml-linearize-attributes attr-list)
 (let ((lgt (length attr-list)))
  ; catch a typical error here: non-symbol key, probably old-fashioned calling convention from html-v1.
  (if (and (>= lgt 1) (not (symbol? (car attr-list))))
      (error (string-append "xml-linearize-attributes: Non-symbol key encountered: " (as-string (car attr-list)) " in attribute list " (as-string attr-list) ". Maybe conversion problem from html-v1.")))
  (xml-linearize-attributes-1 (reverse attr-list) "" "" lgt attr-list)))

; Internal helping operation to xml-linearize-attributes. The parameter whole-attr-list is for error messaging purposes
(define (xml-linearize-attributes-1 attr-list html-attr-string css-attr-string lgt whole-attr-list)
  (cond ((= lgt  0 ) (cons (strip-trailing-characters (list #\space) html-attr-string) css-attr-string))
        ((>= lgt 2) (let* ((val (car attr-list))   ; val and key are reversed at this time
                           (key (cadr attr-list))
                           (css-key (xml-css-key? key))
                         )
                     (cond (css-key                  ; CSS attribute
                              (xml-linearize-attributes-1 
                                (cddr attr-list)
                                html-attr-string 
                                (string-append (xml-linearize-attribute-pair-css val css-key) ";" css-attr-string)
                                (- lgt 2)
                                whole-attr-list))
                           (else                            ; HTML attribute
                             (xml-linearize-attributes-1 
                               (cddr attr-list)
                               (string-append (xml-linearize-attribute-pair-html val key) " " html-attr-string)
                               css-attr-string
                               (- lgt 2)
                               whole-attr-list)))))
        ((< lgt 2) (error (string-append "Xml-Linearize-attributes-1: Called with an odd length attribute list. Not a Lisp property list: " (as-string whole-attr-list))))))

; Return whether key is a css-key, for instance a key like css:k or CSS:k
; A true value is the key propper, such as k (as a string).
(define (xml-css-key? key)
  (let* ((key-str (symbol->string key))
         (lgt (string-length key-str)))
   (if
    (and (>= lgt 4) 
         (or (eq? (string-ref key-str 0) #\c) (eq? (string-ref key-str 0) #\C))
         (or (eq? (string-ref key-str 1) #\s) (eq? (string-ref key-str 1) #\S))
         (or (eq? (string-ref key-str 2) #\s) (eq? (string-ref key-str 2) #\S))
         (eq? (string-ref key-str 3) #\:))
    (substring key-str 4 lgt)
    #f)))

(define (xml-linearize-attribute-pair-html val key)
  (string-append (as-string key) " = " (string-it (as-string val))))

(define (xml-linearize-attribute-pair-css val key)
  (string-append key ": " (as-string val)))





; -----------------------------------------------------------------------------------------------------------------------------
;;; Other useful functions. 
;;; In this section there are non-mirror functions which are useful in in the context of XML and LAML.

;; Define partial attributes and element contents of element and return a new element function
;; with the attributes and contents partially fixed.
;; attributes-and-contents is of the same form as the parameters to a LAML surface mirror functions.
;; In fact, attributes-and-contents are just appended to the actual parameters to the modified.
;; .example (define a-main (modify-element a 'target "main"))
;; .returns an attribute-modified mirror function
(define (xml-modify-element element . attributes-and-contents)
  (lambda parameters (apply element (append parameters attributes-and-contents))))

;; Convert an abtract syntax tree to a parse tree.
;; Abstract syntax trees are produced by the validating mirror functions.
;; Parse trees are used as an internal format in the HTML and the XML pretty printing procedures.
;; You can use the function pretty-print-html-parse-tree on the parse tree returned by ast-to-parse-tree.
;; .pre-condition The LAML tool html-support (for parsing and pretty printing) must be loaded for this function to work.
;; .misc To load html-support: <kbd>(laml-tool-load "xml-html-support/html-support.scm")</kbd>
(define (ast-to-parse-tree ast)
 (let ((pt (ast-to-parse-tree-1 ast)))
   (make-final-parse-tree 'html-tree (list pt))))


; The function doing the real work in the ast to parse tree conversion.
(define (ast-to-parse-tree-1 ast)
 (letrec 
  ((subtree-transform     ; the function doing the overall ast to parse tree transformation
    (lambda (x) 
     (cond ((ast? x) (ast-to-parse-tree-1 x))
	   ((cdata? x) x)
	   ((forced-white-space? x) " ")
	   (else (laml-error "subtree-transform:" "Unknown subtree constituent" (as-string x)))))
    )

   (explicit-space-splicing  ; join explicit spaces, " ", to the left string constituent - to avoid empty lines after pretty printing.
     (lambda (parse-tree)
       (if (tree-entry? parse-tree)
           (let* ((node (root-of-parse-tree parse-tree))
                  (subtrees (subtrees-of-parse-tree parse-tree))
                  (new-subtrees (explicit-space-splicing-lst subtrees '()))
                 )
             (make-parse-tree node new-subtrees))
           parse-tree)
      ))

   (explicit-space-splicing-lst
     (lambda (subtree-list res-lst)
       (cond ((null? subtree-list)
                  (reverse res-lst))
             ((null? (cdr subtree-list))   ; only a single element left - get finished
                (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst)))
             ((and (string? (car subtree-list)) (string? (cadr subtree-list)))   ; interesting case - there are two elements
                (if (equal? (cadr subtree-list) " ")    ; join car and cadr via string-append
                    (explicit-space-splicing-lst (cddr subtree-list) (cons (string-append (car subtree-list) (cadr subtree-list)) res-lst))
                    (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst))))
             (else (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst))))))

   (split-attribute-list  ; split in html and css attributes
     (lambda (attr-lst) (split-attribute-list-1 attr-lst '() ())))

   (split-attribute-list-1  ; iterative helping function
     (lambda (attr-lst html-attr-list css-attr-list)
       (cond ((null? attr-lst) (cons (reverse html-attr-list) (reverse css-attr-list)))
             ((xml-css-key? (car attr-lst)) 
                (split-attribute-list-1 (cddr attr-lst) html-attr-list (cons (cadr attr-lst) (cons (car attr-lst) css-attr-list))))
             (else 
                (split-attribute-list-1 (cddr attr-lst) (cons (cadr attr-lst) (cons (car attr-lst) html-attr-list)) css-attr-list)))))

   (to-css-string     ; linearize all css attributes to a single string
     (lambda (css-attr-list)
       (let ((css-attr-alist (propertylist-to-alist css-attr-list)))
        (string-append
         (list-to-string 
          (map 
           (lambda (css-key-val) (xml-linearize-attribute-pair-css (cdr css-key-val) (xml-css-key? (car css-key-val))))
           css-attr-alist)
          ";") ";"))))

   )
   (let* ((element-name (ast-element-name ast))
 	  (subtrees (ast-subtrees ast))
	  (attr-lst (ast-attributes ast))  ; prop list format
          (attr-list-normal-css (split-attribute-list attr-lst))
          (html-attr-list (car attr-list-normal-css)) ; prop list format
          (css-attr-list (cdr attr-list-normal-css)) ; prop list format
          (attr-lst-result 
            (if (null? css-attr-list)
                html-attr-list
                (cons 'style (cons (to-css-string css-attr-list) html-attr-list))))
	  (kind (ast-kind ast))
   
	 )
     (cond ((eq? kind 'single)
              (make-tag-structure
	       'start-end
	       element-name 
	       attr-lst-result))
           ((eq? kind 'double)
              (explicit-space-splicing
               (make-parse-tree 
		(make-tag-structure
		 'start
		 element-name 
		 attr-lst-result)
		(map subtree-transform subtrees))))
           (else (laml-error "ast-to-parse-tree-1:" "Unknown kind of ast" kind))))))


 


